from .base import MriStudy
from .dwi import DwiStudy
from .bold import BoldStudy
from .t1 import T1Study
from .t2 import T2Study
from .t2star import T2starStudy
# from .bold import BoldStudy
# from .ute import Utetudy
from .epi import EpiSeriesStudy
